#include <unity.h>
#include <kbf/wifi.h>
#include <kbf/rtos.h>
#include <kbf/uart.h>

#include "kbf/net.h"

#define KBF_TEST_MDNS_SSID   "kbf_test_wifi_ap"
#define KBF_TEST_MDNS_PASS   "Pas$w0Rd1337"
#define KBF_TEST_MDNS_MASTER "master"
#define KBF_TEST_MDNS_SLAVE  "slave"
#define KBF_TEST_MDNS_SIGNAL "ready"

using namespace kbf;

TEST_CASE("MAC address functions", "[kbf_net]") {
    uint8_t addr[] = {0, 0, 11, 128, 176, 255};

    auto mac = net::MAC(addr);
    TEST_ASSERT_EQUAL_STRING_MESSAGE("00:00:0B:80:B0:FF", mac.str().c_str(), "ctor / str()");

    net::setMac(mac);

    TEST_ASSERT_EQUAL_STRING_MESSAGE(mac.str().c_str(), net::getMac().str().c_str(), "getMac");
}

TEST_CASE("IP address functions", "[kbf_net]") {
    esp_ip4_addr testAddr = {2952898752};
    static const char *const testIP = "192.168.1.176";

    auto ipFromAddr = net::IP(testAddr);
    TEST_ASSERT_TRUE(ipFromAddr)
    TEST_ASSERT_EQUAL_STRING(testIP, ipFromAddr.str().c_str());
    TEST_ASSERT_EQUAL(0  , ipFromAddr[-1]);
    TEST_ASSERT_EQUAL(192, ipFromAddr[0]);
    TEST_ASSERT_EQUAL(168, ipFromAddr[1]);
    TEST_ASSERT_EQUAL(1  , ipFromAddr[2]);
    TEST_ASSERT_EQUAL(176, ipFromAddr[3]);
    TEST_ASSERT_EQUAL(0  , ipFromAddr[4]);

    auto ipFromString = net::IP(testIP);
    TEST_ASSERT_TRUE(ipFromString)
    TEST_ASSERT_EQUAL_STRING(testIP, ipFromString.str().c_str());

    TEST_ASSERT_TRUE_MESSAGE(ipFromAddr == ipFromString, "operator==")
    TEST_ASSERT_EQUAL_STRING(ipFromAddr.str().c_str(), ipFromString.str().c_str());

    for (const auto &invalidIP : {"", "invalid", "255.255.255.256", "123.123.123.", "0.0.-1.0"}) {
        auto ip = net::IP(invalidIP);
        TEST_ASSERT_FALSE(ip)
        TEST_ASSERT_EQUAL_STRING("", ip.str().c_str());
    }

    for (const auto &validIp : {"0.0.0.0", "255.255.255.255"}) {
        auto ip = net::IP(validIp);
        TEST_ASSERT_TRUE(ip)
        TEST_ASSERT_EQUAL_STRING(validIp, ip.str().c_str());
    }
}

using std::shared_ptr;

static rtos::EventGroup eventGroup;
static net::MDNS        *mdns;

static shared_ptr<kbf::UART> setupUart() {
    auto uart1 = new kbf::UART(1);
    uart1->setPins(33, 32, 25, 26); // TODO use Kconfig
    return shared_ptr<kbf::UART>(uart1);
}

void mdnsMaster() {
    wifi::start(wifi::AP::create(KBF_TEST_MDNS_SSID, KBF_TEST_MDNS_PASS));
    mdns = new net::MDNS();
    mdns->setHostname(KBF_TEST_MDNS_MASTER);

    setupUart()->waitFor(KBF_TEST_MDNS_SIGNAL);

    auto ip = mdns->query(KBF_TEST_MDNS_SLAVE);
    TEST_ASSERT_NOT_NULL(ip)
    TEST_ASSERT_EQUAL_STRING("192.168.4.2", ip->str().c_str());

    delete mdns;
    wifi::stop();
}

void mdnsSlave() {
    wifi::start();
    auto sta = wifi::getSTA();
    eventGroup.clear();

    sta->onIp = {[]() {
        mdns = new net::MDNS(KBF_TEST_MDNS_SLAVE);

        TEST_ASSERT_NULL(mdns->query("nonexistent"))

        auto ip = mdns->query(KBF_TEST_MDNS_MASTER);
        TEST_ASSERT_NOT_NULL(ip)
        TEST_ASSERT_EQUAL_STRING("192.168.4.1", ip->str().c_str());
        setupUart()->write(KBF_TEST_MDNS_SIGNAL);
    }};

    sta->onDisconnect = {[]() {
        eventGroup.setBit(0);
    }};

    sta->connect(KBF_TEST_MDNS_SSID, KBF_TEST_MDNS_PASS);

    eventGroup.waitForBit(0);
    delete mdns;
    wifi::stop();
}

TEST_CASE_MULTIPLE_DEVICES("mDNS", "[kbf_net]", mdnsMaster, mdnsSlave)
